// Try compiling with gcc Beam2D.c -I ~/include -I ~/Mesa-7.0.3/include -L ~/Mesa-7.0.3/lib -lGL -lGLU -lglut -lm -lX11 -lXi -lXmu -o Beam2D (this may now be updated in makefile - OK, g++ is a pain about type conversions)
//#define SBCLIB_GRAPHICS_DX // Link ddraw.lib; otherwise OpenGL32.lib
#include <bitmapfonts.c>
#include <colmap.c>
#include <rnd/Mini.c>
#include <rnd/sq.c>
#include <rnd/cu.c>
#include <rnd/invsq.c>
#include <rnd/setcwdtohomedir.c>
#include <rnd/istouchscreen.c>

void psplots(const double *x,const double *y,const double *xs,const double *ys,const double *np,const int m,
	const char highlight,const double hscale,const double vscale,const double dpp)
{
	int n;
	line(0,yres*3/4,xres/3,yres*3/4,0x808080); // Axes
	line(xres/6,yres/2,xres/6,yres-1,0x808080);
	line(xres*2/3,yres*3/4,xres-1,yres*3/4,0x808080);
	line(xres*5/6,yres/2,xres*5/6,yres-1,0x808080);
	// Some useful statistics on the beam size etc.
	static double bbx=0,bbxs=0,bby=0,bbys=0;
	if (KEY(VK_R)) bbx=bbxs=bby=bbys=0;
	double bx=0,bxs=0,by=0,bys=0;
	for (n=m-1;n>=0;n--)
	{
		bx=Max(bx,x[n]); bxs=Max(bxs,xs[n]);
		by=Max(by,y[n]); bys=Max(bys,ys[n]);
	}
	bbx=Max(bbx,bx); bbxs=Max(bbxs,bxs);
	bby=Max(bby,by); bbys=Max(bbys,bys);
	char str[200];
	sprintf(str,"Max |x| = %lg m (%lg)\nMax |x'| = %lg (%lg)\n\
Max |y| = %lg m (%lg)\nMax |y'| = %lg (%lg)",
		bx,bbx,bxs,bbxs,by,bby,bys,bbys);
	bfsetfont("ncgismall");
	bfwrite(xres*0.45,yres*3/4,str,GFX_white);
	bfsetfont("ncgibig");
	#ifdef SBCLIB_GRAPHICS_OPENGL
	glBegin(GL_POINTS);
	unsigned c;
	#endif
	for (n=m-1;n>=0;n--)
	{
		float xx=x[n]*hscale+xres/6,yy=xs[n]*-vscale+yres*3/4;
		if (xx<=xres/3 && yy>=yres/2) // X-axis phase space
		{
			c=(dpp>0?colmap5(np[n],0.999-dpp*0.9,1.001+dpp*0.9):GFX_white);
			#ifdef SBCLIB_GRAPHICS_OPENGL
			glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF); glVertex2f(xx,yy);
			#else
			putpix(xx,yy,c);
			#endif
			if (n==0 && highlight)
			{
				float th,fx,fy;
				for (th=0;th<M_TWOPI;th+=0.1)
				{
					fx=cos(th)*5; fy=sin(th)*5;
					#ifdef SBCLIB_GRAPHICS_OPENGL
					glColor3f(1,0,1); glVertex2f(xx+fx,yy+fy);
					#else
					putpix(xx+fx,yy+fy,RGB666(5,0,5));
					#endif
				}
			}
		}
		xx=y[n]*hscale+xres*5/6; yy=ys[n]*-vscale+yres*3/4;
		if (xx>=xres*2/3 && yy>=yres/2) // Y-axis phase space
		{
			c=(dpp>0?colmap5(np[n],0.999-dpp*0.9,1.001+dpp*0.9):GFX_white);
			#ifdef SBCLIB_GRAPHICS_OPENGL
			glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF); glVertex2f(xx,yy);
			#else
			putpix(xx,yy,c);
			#endif
			if (n==0 && highlight)
			{
				float th,fx,fy;
				for (th=0;th<M_TWOPI;th+=0.1)
				{
					fx=cos(th)*5; fy=sin(th)*5;
					#ifdef SBCLIB_GRAPHICS_OPENGL
					glColor3f(1,0,1); glVertex2f(xx+fx,yy+fy);
					#else
					putpix(xx+fx,yy+fy,RGB666(5,0,5));
					#endif
				}
			}
		}
	}
	#ifdef SBCLIB_GRAPHICS_OPENGL
	glEnd();
	#endif
}

void plotit(const double *x,const double *y,const int m,const double scale,const char phasespace)
{
	int n,cy=(phasespace?yres*0.36:yres/2);
	#ifdef SBCLIB_GRAPHICS_OPENGL
	ccl();
	glBegin(GL_POINTS);
	glColor3f(1,1,1);
	for (n=0;n<m;n++) // Plotting loop
		glVertex2d(x[n]*scale+xres/2,y[n]*scale+cy);
	glEnd();
	ccl();
	#else
	cl(0);
	for (n=0;n<m;n++) // Plotting loop
		putpix(x[n]*scale+xres/2,y[n]*scale+cy,GFX_white);
	update();
	#endif
}

void helpscreen(void)
{
	#ifdef SBCLIB_GRAPHICS_OPENGL
	ccl();
	#else
	cl(0);
	#endif
	bfwrite(100,90,"Key list:",GFX_white);
	bfwrite(100,120,"P,K,W,G: Create matched Parallel, K-V, Waterbag or Gaussian beam\n\
(also resets simulation)\n\
\n\
2,3: Left/Right Steering Dipoles\n\
\n\
4,5: Focussing/Defocussing Quadrupoles\n\
\n\
6,7: Sextupoles\n\
\n\
8,9: Octupoles\n\
\n\
0,O: Apply solenoid field\n\
\n\
D: Toggle momentum spread\n\
\n\
F: Toggle FODO cell with matched beam\n\
\n\
I: Toggle ISIS lattice with matched beam\n\
\n\
+: Increase simulation speed\n\
-: Decrease simulation speed\n\
\n\
H: Toggle highlighting of one particle\n\
\n\
A: Toggle phase-space graphs\n\
\n\
L: Toggle luminous particles\n\
\n\
S: Toggle scraper\n\
left/right: Move scraper\n\
\n\
C: Toggle space-charge (uniform beam approximation)\n\
\n\
Q/Escape: Quit",RGB666(0,5,5));
	bfwrite(100,500,"Press space to continue...",RGB666(0,0,5));
	#ifdef SBCLIB_GRAPHICS_OPENGL
	ccl();
	#endif
	KEYUP(VK_F1);
	do
	{
		#ifndef SBCLIB_GRAPHICS_OPENGL
		update();
		#endif
	}
	while (!dealwithmessages() && !KEY(VK_SPACE));
}

#define L 500

char name[L][16];
double len[L],kval[L],rho[L],rigidity=1.235293713;
int l,superperiods=10;

void getlattice(const char *filename)
{
	FILE *in=fopen(filename,"rt");
	for (l=0;!feof(in);l++)
	{
		fscanf(in,"%s %lf %lf %lf\n",name[l],&len[l],&kval[l],&rho[l]);
		kval[l]/=rigidity;
		if (rho[l]) rho[l]=len[l]/rho[l];
	}
	fclose(in);
}

void drawdiamond(const int x,const int y,const int s,const unsigned c)
{
	line(x-s,y,x+s,y,c);
	int n; for (n=s;n>0;n--)
	{
		line(x-s+n,y-n,x+s-n,y-n,c);
		line(x-s+n,y+n,x+s-n,y+n,c);
	}
}

void plotlattice(const unsigned c,const unsigned cpos,const unsigned cneg,const unsigned ch,double s)
{
	int sp,n;
	double splength=0; for (n=0;n<l;n++) splength+=len[n];
	s=fmod(s,splength*superperiods);
	static double ox=0,oy=0,oangle=0; const double cx=xres/4+0.5,cy=yres/3+0.5;
	double px=cx-ox,py=cy-oy,bx,by,angle=-oangle,scale=xres/100;
	#ifdef SBCLIB_GRAPHICS_OPENGL
	glBegin(GL_LINES);
	#define LI line_inner
	#else
	#define LI line
	#endif
	for (sp=0;sp<superperiods;sp++) for (n=0;n<l;n++)
	{
		if (s>=0 && s<len[n])
		{
			if (!rho[n]) {bx=px-sin(angle)*s*scale; by=py+cos(angle)*s*scale;}
			else
			{
				bx=px+(-cos(angle)+cos(angle+s/rho[n]))*rho[n]*scale;
				by=py+(-sin(angle)+sin(angle+s/rho[n]))*rho[n]*scale;
			}
			ox+=bx-cx; oy+=by-cy; oangle+=angle*0.01;
		}
		s-=len[n];
		float opx=px,opy=py,oa=angle,npx,npy;
		if (!rho[n]) {px-=sin(angle)*len[n]*scale; py+=cos(angle)*len[n]*scale; LI(opx,opy,px,py,c);}
		else
		{
			float a,da=1.0/rho[n]/scale,ang=len[n]/rho[n];
			for (a=da;a<ang+da;a+=da)
			{
				npx=opx+(-cos(angle)+cos(angle+a))*rho[n]*scale;
				npy=opy+(-sin(angle)+sin(angle+a))*rho[n]*scale;
				LI(px,py,npx,npy,c);
				px=npx; py=npy;
			}
			px=opx+(-cos(angle)+cos(angle+ang))*rho[n]*scale;
			py=opy+(-sin(angle)+sin(angle+ang))*rho[n]*scale;
			angle+=ang;
		}
		if (kval[n])
		{
			float pr=0.8*scale; unsigned col=(kval[n]>0?cpos:cneg);
			LI(opx-pr*cos(oa),opy-pr*sin(oa),opx+pr*cos(oa),opy+pr*sin(oa),col);
			LI(opx+pr*cos(oa),opy+pr*sin(oa),px+pr*cos(angle),py+pr*sin(angle),col);
			LI(px+pr*cos(angle),py+pr*sin(angle),px-pr*cos(angle),py-pr*sin(angle),col);
			LI(px-pr*cos(angle),py-pr*sin(angle),opx-pr*cos(oa),opy-pr*sin(oa),col);
			pr+=fabs(kval[n])*scale;
			LI(opx-pr*cos(oa),opy-pr*sin(oa),opx+pr*cos(oa),opy+pr*sin(oa),col);
			LI(opx+pr*cos(oa),opy+pr*sin(oa),px+pr*cos(angle),py+pr*sin(angle),col);
			LI(px+pr*cos(angle),py+pr*sin(angle),px-pr*cos(angle),py-pr*sin(angle),col);
			LI(px-pr*cos(angle),py-pr*sin(angle),opx-pr*cos(oa),opy-pr*sin(oa),col);
		}
	}
	#ifdef SBCLIB_GRAPHICS_OPENGL
	glEnd();
	#endif
	#undef LI
	drawdiamond(bx,by,2,ch);
}

#include <fonts/lets.dat.c>
#include <fonts/ters.dat.c>
#include <fonts/captionfont.dat.c>
#include "decorativeparticles.c"
#include "decorativequarks.c"

void plotpipe(const float yorig,const float pipe_radius,const float scale)
{
	//circle(xres/2,yorig,pipe_radius*scale,0x0000FF);
	int n; float ox=xres/2,oy=yorig,r=pipe_radius*scale,th;
	glBegin(GL_TRIANGLE_FAN);
	glColor3f(0,0,0);
	glVertex2f(ox,oy);
	glColor3f(0.15,0.15,0.15);
	for (n=0;n<=100;n++)
	{
		th=M_TWOPI*n/100;
		glVertex2f(ox+r*cos(th),oy+r*sin(th));
	}
	glEnd();
	glBegin(GL_QUAD_STRIP);
	glColor3f(0.5,0.5,0.5); float ri=r*1.02;
	for (n=0;n<=100;n++)
	{
		th=M_TWOPI*n/100;
		glVertex2f(ox+r*cos(th),oy+r*sin(th));
		glVertex2f(ox+ri*cos(th),oy+ri*sin(th));
	}
	glEnd();
	/*// Old version
	#ifdef SBCLIB_GRAPHICS_OPENGL
	glBegin(GL_POINTS); glColor3f(0,0,1);
	#endif
	for (n=0;n<1000;n++) // The pipe
	{
		float th=M_TWOPI*n/1000,fx,fy;
		fx=pipe_radius*sin(th); fy=pipe_radius*cos(th);
		#ifdef SBCLIB_GRAPHICS_OPENGL
		glVertex2f(fx*scale+xres/2,fy*scale+yorig);
		#else
		putpix(fx*scale+xres/2,fy*scale+yorig,9);
		#endif
	}
	#ifdef SBCLIB_GRAPHICS_OPENGL
	glEnd();
	#endif*/
}

void plotmpole(const float yorig,const float pipe_radius,const float scale,
	const int poles,const float strength)
{
	int n,i; float ox=xres/2,oy=yorig,th,
		r=pipe_radius*scale,ri=r*1.06,ro=r*3,
		ths=M_TWOPI/poles,thh=ths/3;
	r*=1.03;
	bfsetfont("captionfont");
	for (i=poles-1;i>=0;i--)
	{
		if ((i&1)^(strength>0)) glColor3f(1,0.2,0.2); else glColor3f(0.2,0.6,1);
		glBegin(GL_QUAD_STRIP);			
		for (n=0;n<=20;n++)
		{
			th=ths*(0.5+i)+thh*(n-10)*0.1;
			glVertex2f(ox+r*cos(th),oy+r*sin(th));
			glVertex2f(ox+ri*cos(th),oy+ri*sin(th));
		}
		glEnd();
		glColor3f(0.2,0.2,0.2);
		glBegin(GL_QUAD_STRIP);			
		for (n=0;n<=20;n++)
		{
			th=ths*(0.5+i)+thh*(n-10)*0.1;
			glVertex2f(ox+ri*cos(th),oy+ri*sin(th));
			glVertex2f(ox+ro*cos(th),oy+ro*sin(th));
		}
		glEnd();
		th=ths*(0.5+i);
		bfwrite(ox+r*1.1*cos(th)-0.5*BF_height,oy+r*1.1*sin(th)-0.5*BF_height,(i&1)^(strength>0)?"N":"S",GFX_white);
	}
	bfsetfont("ncgibig");
}

void demohelp(void)
{
	bfsetfont("captionfont");
	bfwrite(0,0,"Particle Beam Focussing Game",GFX_white);
	bfwrite(0,BF_height*2,"Goal: to make the beam as small\n\
as possible to create collisions.\n\
\n\
The left/right arrow keys apply\n\
'quadrupole' magnets to change\n\
the beam shape.\n\
\n\
'Luminosity' of collisions is shown\n\
at the right of the screen.\n\
\n\
Press Space Bar to reset.",0x80C080);
	bfsetfont("ncgibig");
}

#include <resource.c>
#include "isis_sp.txt.c"

WINMAIN
{
	int n,i,k,m=0,M=10000; char str[100]; FILE *in;
	double *x=(double *)malloc(M*sizeof(double)),*y=(double *)malloc(M*sizeof(double)),
		*xs=(double *)malloc(M*sizeof(double)),*ys=(double *)malloc(M*sizeof(double)),
		*np=(double *)malloc(M*sizeof(double)); // Normalised momentum = 1 + dp/p
	intel_hd_graphics_cludge=1;
	GRAPHICS(0,0,"2D Beam Simulation [Stephen Brooks 2002-15]");
	loadfontresource(ters_dat,"ncgibig");
	loadfontresource(lets_dat,"ncgismall");
	loadfontresource(captionfont_dat,"captionfont");
	double pipe_radius=0.1,width=1.5*pipe_radius*xres/yres,scraper_x=0.03,dpp=0,splength,s=0,ds=0.01,mps=1;
	pipe_radius=0.07;
	char fodo=0,isis=0,oisis=0,highlight=0,phasespace=0,spacecharge=0,scraper=0;
	setcwdtohomedir();
	//in=fopen("res.txt","rt"); fscanf(in,"%d x %d",&xx,&yy); fclose(in);
	unpack_resource(isis_sp_txt,"isis_sp.txt",None);
	getlattice("isis_sp.txt");
	splength=0; for (n=0;n<l;n++) splength+=len[n];
	#ifdef DECORATIVEQUARKS
	LSet quarks=LSet(Quark);
	#endif
	int decorativeparticles=0; GLuint ptex=protontex(32); float proton_r=16;
	double peaklum=1e-6;
	char new_beam,bend,touch=istouchscreen(); int omx=mx,omy=my,omb=mb;
	LOOP
	{
		start:;
		double framestart=eithertime(),scale=(double)xres/(width*(phasespace?1.5:1)),remnant=1e20;
		double p2=(double)(KEY(VK_2)-KEY(VK_3))/30,
			p4=2*(KEY(VK_4)-KEY(VK_5)),
			p6=(KEY(VK_6)-KEY(VK_7))*30,
			p8=(KEY(VK_8)-KEY(VK_9))*30*30,
			s0=(KEY(VK_O)-KEY(VK_0));
		p4+=2*(KEY(VK_LEFT)-KEY(VK_RIGHT));
		if (touch) if (mb && omb) p4-=((double)mx-omx)/(xres/250);//(mx>omx?2:mx<omx?-2:0);
		int yorig=(phasespace?yres*0.36:yres/2);
		if (p2) plotmpole(yorig,pipe_radius,scale,2,p2);
		if (p4) plotmpole(yorig,pipe_radius,scale,4,p4);
		if (p6) plotmpole(yorig,pipe_radius,scale,6,p6);
		if (p8) plotmpole(yorig,pipe_radius,scale,8,p8);
		new_beam=0;
		if (isis)
		{
			bfsetfont("ncgismall");
			double rs=fmod(s,splength),rt=0;
			plotlattice(0x800080,0x800000,0x008080,GFX_white,s);
			for (k=0;rt<=rs;k++) rt+=len[k]; k--;
			remnant=rt-rs;
			p4=kval[k]; bend=(rho[k]!=0);
			sprintf(str,"Component #%d %s",k,name[k]);
			bfwrite(xres/4,20,str,GFX_white);
			sprintf(str,"Simulation speed %.2f m/s",mps);
			bfwrite(xres/4,40,str,GFX_white);
			sprintf(str,"Distance around ring %.2f m",fmod(s,splength*superperiods));
			bfwrite(xres/4,50,str,GFX_white);
			sprintf(str,"Superperiod %d / %d",(int)floor(s/splength)%superperiods+1,superperiods);
			bfwrite(xres/4,60,str,GFX_white);
			bfsetfont("ncgibig");
		}
		else k=-1;
		if (KHIT(VK_F1)) helpscreen();
		if (KHIT(VK_ADD)) mps*=2;
		if (KHIT(VK_SUBTRACT)) mps/=2;
		int newbeamtype=0;
		if (KHIT(VK_P)) newbeamtype=1; // Parallel beam
		else if (KHIT(VK_K) || isis && !oisis) newbeamtype=2; // Non-Unit K-V beam
		else if (KHIT(VK_W)) newbeamtype=3; // Non-Unit waterbag
		else if (KHIT(VK_G)) newbeamtype=4; // Gaussian
		else if (KHIT(VK_SPACE) || !m) newbeamtype=5; // Almost parallel beam (waterbag with small divergence)
		if (newbeamtype)
		{
			m=M;
			for (n=0;n<m;n++)
			{
				switch (newbeamtype)
				{
					case 1:
					{				
						do {x[n]=frnd(2)-1; y[n]=frnd(2)-1;}
						while (sq(x[n])+sq(y[n])>1);
						xs[n]=ys[n]=0;
					} break;
					case 2:
					{
						float dd;
						do {x[n]=frnd(2)-1; y[n]=frnd(2)-1; xs[n]=frnd(2)-1; ys[n]=frnd(2)-1;}
						while ((dd=sq(x[n])+sq(y[n])+sq(xs[n])+sq(ys[n]))>1);
						dd=sqrt(dd);
						x[n]/=dd; y[n]/=dd; xs[n]/=dd; ys[n]/=dd;
					} break;
					case 3: case 5:
					{
						float dd;
						do {x[n]=frnd(2)-1; y[n]=frnd(2)-1; xs[n]=frnd(2)-1; ys[n]=frnd(2)-1;}
						while ((dd=sq(x[n])+sq(y[n])+sq(xs[n])+sq(ys[n]))>1);
					} break;
					case 4:
					{
						twonormalvariates(&x[n],&y[n]);
						twonormalvariates(&xs[n],&ys[n]);
						x[n]/=4; y[n]/=4; xs[n]/=4; ys[n]/=4;
					} break;
				}
				if (isis)
				{
					x[n]*=0.0323835344; xs[n]*=-0.002313357742;
					y[n]*=0.02434880285; ys[n]*=0.004573547238;
				}
				else if (newbeamtype==5) {x[n]/=20; y[n]/=20; xs[n]/=1500; ys[n]/=1500;}
				else {x[n]/=25; y[n]/=25; xs[n]/=75; ys[n]/=75;}
				np[n]=1.0+(frnd(2)-1)*dpp;
			}
			new_beam=1; oisis=isis;
		}
		#ifdef DECORATIVEQUARKS
		if (new_beam) LSet_clear(&quarks);
		#endif
		if (new_beam) peaklum=1e-6;
		if (new_beam && s!=0) {s=1e-6; dealwithmessages(); goto start;}
		if (KHIT(VK_D)) // Toggle momentum spread
		{
			dpp=0.15-dpp;
			for (n=m-1;n>=0;n--) np[n]=1.0+(frnd(2)-1)*dpp;
		}
		oisis=isis;
		if (KHIT(VK_I)) // Toggle ISIS
		{
			isis^=1; if (isis) fodo=0; else mps=1;
		}
		if (KHIT(VK_F)) // Toggle FODO cell
		{
			fodo^=1;
			if (fodo)
			{
				isis=0;
				m=M; s=0;
				for (n=0;n<m;n++)
				{
					float dd;
					do {x[n]=frnd(2)-1; y[n]=frnd(2)-1; xs[n]=frnd(2)-1; ys[n]=frnd(2)-1;}
					while ((dd=sq(x[n])+sq(y[n])+sq(xs[n])+sq(ys[n]))>1);
					dd=sqrt(dd);
					x[n]/=dd; y[n]/=dd; xs[n]/=dd; ys[n]/=dd;
					x[n]*=2.2882/40; y[n]*=1.43185/40;
					xs[n]/=2.2882*40; ys[n]/=1.43185*40;
				}
			}
			else m=0;
		}
		if (fodo)
		{
			i=floor(fmod(s,3)*4);
			if (i==0 || i==11) p4+=1.384;
			else if (i==5 || i==6) p4-=1.384;
		}
		if (KHIT(VK_H)) // Highlight a particle
		{
			highlight^=1;
			if (highlight)
			{
				double tx=x[0],ty=y[0],txs=xs[0],tys=ys[0];
				i=rnd(m);
				x[0]=x[i]; y[0]=y[i]; xs[0]=xs[i]; ys[0]=ys[i];
				x[i]=tx; y[i]=ty; xs[i]=txs; ys[i]=tys;
			}
		}
		if (KHIT(VK_A)) phasespace^=1; // Toggle phase-space graphs
		if (KHIT(VK_L)) decorativeparticles^=1; // Toggle luminous particles
		if (KHIT(VK_S) && !KEY(VK_CONTROL)) scraper^=1; // Toggle scraper
		if (KHIT(VK_C)) spacecharge^=1; // Toggle space-charge
		if (spacecharge)
		{
			double a=0,b=0,Q=0.0005,c;
			for (n=0;n<m;n++)
			{
				a=Max(a,fabs(x[n]));
				b=Max(b,fabs(y[n]));
			}
			c=Q/(a+b);
			for (n=0;n<m;n++)
			{
				xs[n]+=c*x[n]/a*ds/np[n];
				ys[n]+=c*y[n]/b*ds/np[n];
			}
		}
		double txx=0,tyy=0;
		for (n=0;n<m;n++) // Advance particles a distance ds
		{
			xs[n]-=(p2+p4*x[n]+p6*(sq(x[n])-sq(y[n]))+p8*(cu(x[n])-x[n]*sq(y[n])*3)-s0*ys[n])*ds/np[n];
			ys[n]+=(p4*y[n]+p6*x[n]*y[n]*2+p8*(sq(x[n])*y[n]*3-cu(y[n]))-s0*xs[n])*ds/np[n];
			if (bend && k>=0) xs[n]-=invsq(rho[k])*x[n]*ds/np[n];
			x[n]+=xs[n]*ds; y[n]+=ys[n]*ds;
			if (sq(x[n])+sq(y[n])>sq(pipe_radius) || scraper && x[n]>scraper_x)
			{
				m--;
				x[n]=x[m]; y[n]=y[m]; xs[n]=xs[m]; ys[n]=ys[m]; np[n]=np[m];
				if (n==0) highlight=0;
				n--;
			}
			txx+=sq(x[n]); tyy+=sq(y[n]);
		}
		txx/=m; tyy/=m;
		plotpipe(yorig,pipe_radius,scale);
		// Plotting loop
		{
			int cy=(phasespace?yres*0.36:yres/2);
			#ifdef SBCLIB_GRAPHICS_OPENGL
			glBegin(GL_LINES);
			#define LI line_inner
			#else
			#define LI line
			#endif
			for (n=m-1;n>=0;n--)
			{
				LI(x[n]*scale+xres/2,y[n]*scale+cy,
					(x[n]-xs[n]*ds*2)*scale+xres/2,(y[n]-ys[n]*ds*2)*scale+cy,(dpp>0?colmap5(np[n],0.999-dpp*0.9,1.001+dpp*0.9):GFX_white));
			}
			#ifdef SBCLIB_GRAPHICS_OPENGL
			glEnd();
			#endif
			#undef LI
			#ifdef SBCLIB_GRAPHICS_OPENGL
			// Plotting endpoints
			float r[256],g[256],b[256]; paldata666(r,g,b); int c;
			if (decorativeparticles)
			{
				glEnable(GL_TEXTURE_2D); glBindTexture(GL_TEXTURE_2D,ptex);
				glEnable(GL_BLEND); glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_COLOR);
				glBegin(GL_QUADS);
			}
			else glBegin(GL_POINTS);
			glColor3f(1,1,1);
			for (n=m-1;n>=0;n--) // Plotting loop
			{
				if (dpp>0)
				{
					c=colmap5(np[n],0.999-dpp*0.9,1.001+dpp*0.9);
					glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
				}
				if (decorativeparticles) texquad(x[n]*scale+xres/2,y[n]*scale+cy,proton_r);
				else glVertex2d(x[n]*scale+xres/2,y[n]*scale+cy);
			}
			glEnd();
			if (decorativeparticles) {glDisable(GL_TEXTURE_2D); glDisable(GL_BLEND);}
			#endif
			if (highlight)
			{
				float th,fx,fy;
				#ifdef SBCLIB_GRAPHICS_OPENGL
				glBegin(GL_POINTS); glColor3f(1,0,1);
				#endif
				for (th=0;th<M_TWOPI;th+=0.1)
				{
					fx=cos(th)*5; fy=sin(th)*5;
					#ifdef SBCLIB_GRAPHICS_OPENGL
					glVertex2d((x[0]-xs[0]*ds)*scale+xres/2+fx,(y[0]-ys[0]*ds)*scale+cy+fy);
					#else
					putpix((x[0]-xs[0]*ds)*scale+xres/2+fx,(y[0]-ys[0]*ds)*scale+cy+fy,RGB666(5,0,5));
					#endif
				}
				#ifdef SBCLIB_GRAPHICS_OPENGL
				glEnd();
				#endif
			}
		}
		if (scraper)
		{
			int px=scraper_x*scale+xres/2,dy=sqrt(sq(pipe_radius)-sq(scraper_x))*scale;
			line(px,yorig-dy,px,yorig+dy,0x800000);
			if (KEY(VK_LEFT)) scraper_x=noneg(scraper_x-0.0005);
			if (KEY(VK_RIGHT)) scraper_x=Min(scraper_x+0.0005,pipe_radius);
		}
		if (phasespace) psplots(x,y,xs,ys,np,m,highlight,scale,scale*2,dpp); // Phase-space graphs
		#ifdef DECORATIVEQUARKS
		Quarks_advanceplot(&quarks);
		#endif
		if (fodo) bfwrite(xres/2,0,"FODO cell",RGB666(0,5,5));
		if (spacecharge) bfwrite(xres/2,0,"Space-Charge",RGB666(0,5,5));
		if (isis && k>=0)
		{
			sprintf(str,"Bend Radius: %.2f",rho[k]);
			bfwrite(xres/4,0,str,(bend?RGB666(0,5,0):0x008000));
		}
		sprintf(str,"Quadrupole: %.2f",p4);
		bfwrite(xres*3/4,0,str,(p4!=0?RGB666(0,5,0):0x008000));
		sprintf(str,"Sextupole: %.2f",p6);
		bfwrite(xres*3/4,10,str,(p6!=0?RGB666(0,5,0):0x008000));
		sprintf(str,"Octupole: %.2f",p8);
		bfwrite(xres*3/4,20,str,(p8!=0?RGB666(0,5,0):0x008000));
		sprintf(str,"Solenoid: %.2f",s0);
		bfwrite(xres*3/4,35,str,(s0!=0?RGB666(0,5,0):0x008000));
		double lum=3e27/sqrt(Max(txx,1e-16)*Max(tyy,1e-16))*m/M; int lpot=floor(log10(lum)); unsigned c;
		if (lum>peaklum) peaklum=lum;
		if (lpot<=30) c=0x808000;
		else if (lpot==31) c=0xC0C000;
		else if (lpot==32) c=0xFFF000;
		else if (lpot==33) c=0xFFFF80;
		else if (lpot>=34) c=0xFFFFFF;
		//sprintf(str,"Luminosity: %lg",lum);
		demohelp();
		//bfsetfont("captionfont:scale=150%");
		bfsetfont(":scale=2");
		sprintf(str,"Luminosity: %.1lf x 10^{%d}",lum/pow(10,lpot),lpot); eqnencode(str);
		bfwrite(xres*3/4,50,str,c);
		lpot=floor(log10(peaklum));
		sprintf(str,"Peak: %.1lf x 10^{%d}",peaklum/pow(10,lpot),lpot); eqnencode(str);
		bfwrite(xres*3/4,50+BF_height*1.5,str,0x0000FF);
		if (m<M)
		{
			sprintf(str,"BEAM LOSS: %.2f%%",100.0*(1.0-(float)m/M)); 
			bfwrite(xres*3/4,50+BF_height*3,str,0xFF0000);
		}
		if (touch)
		{
			if (mx<bflength("Reset")+BF_height && my>=50-BF_height && my<50+BF_height*2)
			{
				m=0;
				#ifndef __linux__
				SetCursorPos(xres/2,yres/2);
				#endif
			}
			bfwrite(xres/2,50,"Reset",m?0x800000:GFX_white);
		}
		bfsetfont("ncgibig");
		//rect(xres*3/4,75,lum/1e31,10+lum/1e33,c);
		box(xres*0.94,yres,xres*0.05-1,yres*-0.8*peaklum/3e33+1,0x0000FF);
		rect(xres*0.94,yres,xres*0.05,yres*-0.8*lum/3e33,c);
		#ifdef DECORATIVEQUARKS
		Quarks_addbaryons(&quarks,lum/1e33);
		#endif
		bfwrite(0,yres-BF_height,"Press F1 for help",0x808080);
		if (KHIT(VK_S) && KEY(VK_CONTROL)) screenshotauto("Beam2D");
		omx=mx; omy=my; omb=mb;
		ccl();
		while (eithertime()<framestart+0.012);
		ds=Min(mps*(eithertime()-framestart),remnant+1e-6)*(KEY(VK_SHIFT)?0.2:1);
		s+=ds;
	}
}

