#include <graphics.c>
#include <rnd/sq.c>

GLuint protontex(const int sz)
{
	GLuint ret;
	glGenTextures(1,&ret);
	glBindTexture(GL_TEXTURE_2D,ret); // Selects the current texture
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	int pitch=(sz*3+3)/4*4; // Round up to 4-byte boundary
	unsigned char *d=(unsigned char *)calloc(sz*pitch,1),*p;
	int x,y; float r,th,a;
	for (y=sz-1;y>=0;y--) for (x=sz-1;x>=0;x--) 
	{
		p=d+x*3+y*pitch;
		r=sqrt(sq((float)x-sz/2)+sq((float)y-sz/2)+1e-6)/(sz/2);
		th=atan2((float)y-sz/2,(float)x-sz/2+1e-6);
		a=0.5+Min(0.5,fabs(tan(th*6)));
		if (r>=1) p[0]=p[1]=p[2]=0;
		else p[0]=p[1]=p[2]=(unsigned char)(0xFF*Min(1,a*(0.05/r-0.05)));
	}
	sjbBuild2DMipmaps(3,sz,sz,d);
	free(d);
	return ret;
}

GLuint quarktex(const int sz)
{
	GLuint ret;
	glGenTextures(1,&ret);
	glBindTexture(GL_TEXTURE_2D,ret); // Selects the current texture
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	int pitch=(sz*3+3)/4*4; // Round up to 4-byte boundary
	unsigned char *d=(unsigned char *)calloc(sz*pitch,1),*p;
	int x,y; float r,th,a;
	for (y=sz-1;y>=0;y--) for (x=sz-1;x>=0;x--) 
	{
		p=d+x*3+y*pitch;
		r=sqrt(sq((float)x-sz/2)+sq((float)y-sz/2)+1e-6)/(sz/2);
		th=atan2((float)y-sz/2,(float)x-sz/2+1e-6);
		a=0.5+Min(0.5,fabs(tan(th*6)));
		if (r>=1) p[0]=p[1]=p[2]=0;
		else p[0]=p[1]=p[2]=(unsigned char)(0xFF*Min(1,a*(0.05/r-0.05)));
	}
	sjbBuild2DMipmaps(3,sz,sz,d);
	free(d);
	return ret;
}

void texquad(const float x,const float y,const float r)
{ // Needs glBegin(GL_QUADS) and glEnable(GL_TEXTURE_2D) and glBindTexture etc. and glColor3f
	glTexCoord2f(0,0); glVertex2f(x-r,y-r);
	glTexCoord2f(0,1); glVertex2f(x-r,y+r);
	glTexCoord2f(1,1); glVertex2f(x+r,y+r);
	glTexCoord2f(1,0); glVertex2f(x+r,y-r);
}

/*WINMAIN
{
	int n;
	GRAPHICS(0,0,"Example usage [Stephen Brooks 2014]");
	GLuint ptex=protontex(32);
	LOOP
	{
		glEnable(GL_TEXTURE_2D); glBindTexture(GL_TEXTURE_2D,ptex);
		glBegin(GL_QUADS);
		glColor3f(1,1,1);
		for (n=0;n<20;n++)
		{
			float x=xres/2+yres/3*sin(n),y=yres/2-yres/3*cos(n);
			texquad(x,y,5);
		}
		glEnd();
		glDisable(GL_TEXTURE_2D);
		ccl();
	}
}*/

#define DECORATIVEPARTICLES
