#ifndef SBCLIB_COLMAP
#include <rnd/frnd.c>
#include <rnd/Mini.c>
#include <rnd/Min.c>
#include <rnd/nonegi.c>
#include <rnd/noneg.c>
#include <rnd/sq.c>
#include <rnd/lccoverload.c>

#ifndef SBCLIB_GRAPHICS_OPENGL
#include <bitmapfonts.c> // For RGB macros and BF_and[][]
// 8-bit ones
unsigned colmap8(const double w)
{
	float r,g,b,k=1.0-exp(-fabs(w));
	if (w>=0) {r=0.5+0.5*k; g=k; b=0.5+0.5*k;}
	else {r=0.5-0.5*k; g=k; b=0.5+0.5*k;}
	return RGB666((int)(5.4*r),(int)(5.4*g),(int)(5.4*b));
}

unsigned colmap82(double i,double br=5)
{
	unsigned c; float k=frnd(1);
	if (i>1) i=log(i); else i=0;
	if (i<1) return RGB666(0,i*br+k,0);
	else if (i<2) return RGB666(0,(2.0-i)*br+k,(i-1)*br+k);
	else if (i<3) return RGB666((i-2)*br+k,0,(3-i)*br+k);
	else if (i<4) return RGB666(br,(i-3)*br*0.6+k,0);
	else if (i<5) return RGB666(br,((i-4)*br+k)*0.4+br*0.6,(i-4)*br+k);
	else return RGB666(br,br,br);
}

const unsigned colmap83a[30]={45,51,57,63,69,75,74,73,72,71,
	70,107,144,181,218,15,254,253,252,251,
	250,244,238,232,226,220,221,222,223,224};

unsigned colmap83(const double v,const double v0,const double v1)
{
	if (v<v0) return BF_and[7][colmap83(v0+v0-v,v0,v1)];
	else return colmap83a[nonegi(Mini(29,30.0*(v-v0)/(v1-v0)))];
}

const unsigned colmap8d2[11]={0,
	RGB666(5,0,0),
	RGB666(5,3,0),
	RGB666(5,5,0),
	RGB666(3,5,0),
	RGB666(0,4,0),
	RGB666(0,5,3),
	RGB666(0,5,5),
	RGB666(0,3,5),
	RGB666(0,0,5),
	RGB666(3,0,5)};

const unsigned colmap84a[36]={40,41,42,43,44,45,51,57,63,69,
	75,74,73,72,71,70,106,142,178,214,
	250,244,238,232,226,220,221,222,223,224,
	225,231,237,243,249,255};

unsigned colmap84(const double v,const double v0,const double v1)
{
	if (v<v0) return BF_and[7][colmap84(v0+v0-v,v0,v1)];
	else return colmap84a[nonegi(Mini(35,36.0*(v-v0)/(v1-v0)))];
}

unsigned colmap84d(const double v,const double v0,const double v1)
{
	if (v<v0) return BF_and[7][colmap84d(v0+v0-v,v0,v1)];
	else return colmap84a[nonegi(Mini(35,35.0*(v-v0)/(v1-v0)+frnd(1)))];
}

unsigned greymap8(const double x)
{ // Greys on [0,1]
	return 16+nonegi(Mini(floor(0.5+x*23),23));
}
#endif

// 24-bit ones
unsigned colmap(double x) // Shows small values as well - good for showing up artefacts
{
	if (x==0) return 0x000000;
	if (x<0) {unsigned c=colmap(-x); return (c>>8)|((c&0xFF)<<16);}
	x=log(x)/log(0.1);
	if (x<0) return 0xFFFFFF;
	else if (x<0.5) return 0xFFFF00+(int)(255.0-x*510.0);
	else if (x<1) return 0xFF0000+0x0100*(int)(510.0-x*510.0);
	else if (x<1.5) return 0xFF0000+(int)(x*510.0-510.0);
	else if (x<2) return 0x0000FF+0x010000*(int)(1020.0-x*510.0);
	else if (x<3) return (int)(600.0-x*200.0)+55;
	else if (x<4) return 0x0100*(int)(800.0-x*200.0);
	else if (x<5) return 0x101*(int)(750.0-x*150.0);
	else return (int)(700.0/x*(1.0-fmod(x,1)))+0x100*(int)(2000.0/x/x*(1.0-fmod(x,1)));
}

unsigned colmap2(const double x)
{ // Simple red=positive, blue=negative
	if (x<0) {unsigned q=colmap2(-x); return (q>>16)|(q&0xFF00)|((q&0xFF)<<16);}
	if (x==0) return 0x505050;
	double r=Min(1,pow(x,0.15)),
		b=1.0-Min(1,pow(x,-0.15)),
		g=x/(1.0+x);
	return 0x606060+65536*(int)(r*0x9F)+256*(int)(g*0x9F)+(int)(b*0x9F);
}

unsigned colmap1(double x)
{ // Similar to first one with different handling of negatives
	if (x<0) return 0x808080^colmap(-x);
	x=log(x)/log(0.1)+2.5;
	if (x<0) return 0xFFFFFF;
	else if (x<0.5) return 0xFFFF00+(int)(255.0-x*510.0);
	else if (x<1) return 0xFF0000+0x100*(int)(510.0-x*510.0);
	else if (x<1.5) return 0xFF0000+(int)(x*510.0-510.0);
	else if (x<2) return 0x0000FF+0x10000*(int)(1020.0-x*510.0);
	else if (x<3) return (int)(600.0-x*200.0)+55;
	else if (x<4) return 0x100*(int)(800.0-x*200.0);
	else if (x<5) return 0x101*(int)(750.0-x*150.0);
	else return (int)(700.0/x*(1.0-fmod(x,1)))+0x100*(int)(2000.0/x/x*(1.0-fmod(x,1)));
}

unsigned colmap3(const double n,const double mn,const double mx)
{ // Nice rainbow-map (well sort of a rainbow), for non-small values
	double y=5.0*(n-mn)/(mx-mn); int q;
	if (y<0) return 0x000000;
	if (y<1) {q=(int)(255.9*y); return 0x0000FF+q*0xFF;}
	if (y<2) {q=(int)(255.9*(y-1.0)); return 0x00FF00+q*0x10000;}
	if (y<3) {q=(int)(255.9*(y-2.0)); return 0xFFFF00-q*0x100;}
	if (y<4) {q=(int)(128.0*(y-3.0)); return 0xFF0000-q*0xFFFF;}
	if (y<5) {q=(int)(127.9*(y-4.0)); return 0x800080+q*0x10201;}
	return 0xFFFFFF;
}

unsigned colmap3u(const double x) {return colmap3(x,0,1);}

unsigned colmap5(const double n,const double mn,const double mx)
{ // RGB rainbow(excluding violet)-map
	double y=4.0*(n-mn)/(mx-mn); int q;
	if (y<0) {q=(int)(255.9/(1.0-y)); return q*0x010000;}
	if (y>4) {q=(int)(255.9/(y-3.0)); return q*0x01;}
	if (y<1) {q=(int)(255.9*y); return 0xFF0000+q*0x100;}
	if (y<2) {q=(int)(255.9*(y-1.0)); return 0xFFFF00-q*0x10000;}
	if (y<3) {q=(int)(255.9*(y-2.0)); return 0x00FF00+q*0x01;}
	{q=(int)(255.9*(y-3.0)); return 0x00FFFF-q*0x100;}
}

unsigned colmap4(const double x)
{ // Errr...
	if (x<5) return 0x000000;
	if (x<10) return 0x0000FF;
	if (x<15) return 0x00FFFF;
	if (x<20) return 0x00FF00;
	if (x<25) return 0xFFFF00;
	return 0xFF0000;
}

unsigned greymap(const double x)
{ // Greys on [0,1]
	return 0x010101*nonegi(Mini((int)(0.5+x*255),0xFF));
}

unsigned greymapw(const double x)
{ // The above with warning colours for over/underflow
	if (x<0) return 0x0000FF+0x000100*nonegi(Mini((int)(0.5+(-x)*255),0xFF));
	if (x>1) return 0xFF0000+0x000100*nonegi(Mini((int)(0.5+(x-1)*255),0xFF));
	return 0x010101*nonegi(Mini((int)(0.5+x*255),0xFF));
}

// Complex numbers!
unsigned overloaded colmapcx(const double re,const double im)
{
	const double pfac=1; // Larger values encode more extreme magnitudes better
	double rr=re*re+im*im,th=atan2(im,re)/M_PI*3,mn,mx,fr,fg,fb;
	if (rr<1) {mn=0; mx=pow(rr,0.5/pfac);}
	else {mn=1.0-pow(rr,-0.5/pfac); mx=1;}
	int q=(int)floor(th); float w=mn+(th-q)*(mx-mn),e=mn+mx-w;
	switch (q)
	{
		case -3: case 3: fr=mn; fg=e; fb=mx; break;
		case -2: fr=w; fg=mn; fb=mx; break;
		case -1: fr=mx; fg=mn; fb=e; break;
		case 0: fr=mx; fg=w; fb=mn; break;
		case 1: fr=e; fg=mx; fb=mn; break;
		case 2: fr=mn; fg=mx; fb=w; break;
	}
	return (((unsigned)(fr*255.0))<<16)|
		(((unsigned)(fg*255.0))<<8)|
		((unsigned)(fb*255.0));
}

#ifdef SBCLIB_COMPLEX
unsigned overloaded colmapcx(const Complex z)
{
	return colmapcx(z.re,z.im);
}
#endif

#ifdef SBCLIB_V3
unsigned colmapv(const V3 v,const double mx=1,const int alt=0)
{
	double r=0,g=0,b=0;
	if (alt) {r=noneg(v.x); g=noneg(v.y); b=noneg(v.z);}
	else {r=fabs(v.x); g=fabs(v.y); b=fabs(v.z);}
	return 65536*(int)(Min(1,r/mx)*255.9)+256*(int)(Min(1,g/mx)*255.9)+(int)(Min(1,b/mx)*255.9);
}

unsigned colmapvpow(const V3 v,const double mx=1,const int alt=0,const double p=1)
{ // e.g. p=1.0/2.2 for gamma of 2.2
	double r=0,g=0,b=0;
	if (alt) {r=noneg(v.x); g=noneg(v.y); b=noneg(v.z);}
	else {r=fabs(v.x); g=fabs(v.y); b=fabs(v.z);}
	return 65536*(int)(Min(1,pow(r/mx,p))*255.9)+256*(int)(Min(1,pow(g/mx,p))*255.9)+(int)(Min(1,pow(b/mx,p))*255.9);
}
#endif

// Physics! (see RGBspectrum.xls, http://en.wikipedia.org/wiki/Visible_spectrum and /File:CRT_phosphors.png)
unsigned spectrumcolour1(const double wavelength)
{
	float x=-5.135-13.18*log(wavelength/1e-6);
	if (x<=0 || x>=7) return 0x000000; // Not strictly true since extreme 750nm red is -1.343 and extreme 380nm violet is 7.618
	int i=(int)x,a=(int)(255.999*(x-i));
	switch (i)
	{
		case 0: return a<<16;
		case 1: return 0xFF0000|(a<<8);
		case 2: return 0x00FF00|((0xFF^a)<<16);
		case 3: return 0x00FF00|a;
		case 4: return 0x0000FF|((0xFF^a)<<8);
		case 5: return 0x0000FF|(a<<16);
		case 6: return (0xFF^a)*0x010001;
	}
	return 0x000000;
}

unsigned spectrumcolour(const double wavelength)
{ // Somewhat more realistic version, quadratic slopes with area 2/3 introduced, tweaks to keep colour area constant
	float x=-5.135-13.18*log(wavelength/1e-6);
	if (x<=-1 || x>=8) return 0x000000; // Gives slightly too short red and too long violet ends
	if (x<=1) return ((int)(255.99*sq(0.5*(x+1))))<<16;
	if (x<=2) return 0xFF0000|(((int)(255.99*(x-1)))<<8);
	if (x<=3) return 0x00FF00|(((int)(255.99*(3-x)))<<16);
	if (x<=4) return 0x00FF00|((int)(255.99*(1-sq(0.5*(5-x)))));
	if (x<=5) return (((int)(255.99*(5-x)))<<8)|((int)(255.99*(1-sq(0.5*(5-x)))));
	if (x<=6) return 0x0000FF|(((int)(182.85*(x-5)))<<16); // 182.85 = 255.99*5/7
	return (((int)(182.85*sq(0.5*(8-x))))<<16)|((int)(255.99*sq(0.5*(8-x))));
}

#define SBCLIB_COLMAP
#endif
