#ifndef SBCLIB_GFX_OPENGL
// The OS-specific part of the interface with OpenGL
// Windows part based on tutoral http://www.essi.fr/~buffa/cours/synthese_image/DOCS/Tutoriaux/Nehe/lesson%d.htm
// Includes making the window (Linux part uses SDL), setting the mode, res etc.
// Link OpenGL32.lib for Windows
/* Link -lSomeOtherStuff for Linux, example makefile below
NAME=exp

default:
	# echo -e "\n\n\n\n\n\n\n\n\n\n"
	g++ $(NAME).c -I ~/include -I ~/Mesa-7.0.3/include -L ~/Mesa-7.0.3/lib -lGL -lGLU -lSDL -lm -lX11 -lXi -lXmu -lrt -o $(NAME)
*/
#ifdef __APPLE__
#define __linux__
#endif
#ifndef __linux__
#include <windows.h>	// Header File For Windows
#endif
#include <rnd/macros.c> // Define WINMAIN, DEBUG etc.; KEY, LOOP are in this file
#include <rnd/Mini.c>
#include <rnd/Maxi.c>
#include <GL/gl.h>		// Header File For The OpenGL32 Library
#include <rnd/sjbglgetprocaddress.c>

// Globals interfacing with the higher-level stuff
int xres,yres;
const int keystate_m=512;
unsigned char keystate[keystate_m],oldkeystate[keystate_m];	// Array Used For The Keyboard Routine
int mx=0,my=0,mb=0,omb=0;
char gotch[500]="";
#ifdef SBCLIB_GRAPHICS_WINDOWED
char gl_windowed=1;
#else
char gl_windowed=0;
#endif
int makethewindow_shiftx=0,makethewindow_shifty=0; // I'm *guessing* the primary display starts at (0,0)
#include <report.c> // Needs to know gl_windowed if SBCLIB_GRAPHICS is defined

void ReSizeGLScene(const int xs,int ys); // Generic functions needed from <graphics_opengl.c>
void InitGL(const int xs,const int ys);

#include <rnd/sjbdatetime.c>
void gerrorlog(const char *message)
{ // Called by gerror, which depends on the OS
	FILE *out=fopen("graphics_error.log","at");
	fprintf(out,"%s %s\n",sjbdatetime(),message);
	fclose(out);
}

// Macros identical between Linux and Windows
#define KEYUP(x) while (!dealwithmessages() && KEY(x));

#define LOOP if (!xres) reportfatal("GRAPHICS() macro not called at LOOP start, or no resolution specified"); \
	for (char leave=(dealwithmessages(),0);!leave;leave|=dealwithmessages()|KEY(VK_Q)|KEY(VK_ESCAPE))

#define LOOPESC char leave; for (leave=(dealwithmessages(),0);!leave;leave|=dealwithmessages()|KEY(VK_ESCAPE))
#define PAUSE for (dealwithmessages();!dealwithmessages() && !KHIT(VK_SPACE););

#ifdef __linux__
// Now trying this using SDL
#include <SDL2/SDL.h>
#include <SDL2/SDL_opengl.h>
#include <rnd/sjbshowcursor.c>

SDL_Window *sdl_window=NULL;
SDL_GLContext sdl_gl_context=NULL;

#define GRAPHICS(X,Y,T) if (!makethewindow(X,Y,T)) return;

int makethewindow(int xs,int ys,const char *title)
{ // Returns 1 on success, 0 on failure
	if (SDL_Init(SDL_INIT_EVERYTHING)<0) return 0;
	int wx=0,wy=0;
	if (xs==0) // Autodetect
	{
		if (ys==0) // Primary display
		{
			SDL_Rect r;
			SDL_GetDisplayBounds(0,&r);
			wx=r.x; wy=r.y; xs=r.w; ys=r.h;
		}
		else // Display number ys (starting from 1)
		{
			int dm=SDL_GetNumVideoDisplays(),
				i=(ys>=1 && ys-1<dm?ys-1:0);
			SDL_Rect r;
			SDL_GetDisplayBounds(i,&r);
			wx=r.x; wy=r.y; xs=r.w; ys=r.h;
		}
		if (gl_windowed==1) {ys=(int)(ys*0.6); xs=ys*4/3;}
	}
	else if (xs==-1) // Multi-monitor spanning mode
	{ // (ys<=0) gives all monitors, or span subset of monitors with bit flags in ys
		int dm=SDL_GetNumVideoDisplays(),
		wx=wy=999999999; int vx=-wx,vy=-wy;
		for (int n=0;n<dm;n++) if (ys<=0 || n<31 && (ys&(1<<n)))
		{
			SDL_Rect r;
			SDL_GetDisplayBounds(n,&r);
			if (r.x<wx) wx=r.x;
			if (r.y<wy) wy=r.y;
			if (r.x+r.w>vx) vx=r.x+r.w;
			if (r.y+r.h>vy) vy=r.y+r.h;
		}
		xs=vx-wx;
		ys=vy-wy;
	}
	//if (xs==0) // SDL1
	//{
		//const SDL_VideoInfo *vi=SDL_GetVideoInfo();
		//xs=vi->current_w; ys=vi->current_h;
		//if (gl_windowed) {xs=1024; ys=768;} else {xs=1920; ys=1080;}
	//}
	//if (!SDL_SetVideoMode(xs,ys,32,SDL_OPENGL|(gl_windowed?SDL_RESIZABLE:SDL_FULLSCREEN))) return 0; // SDL1
	SDL_GL_SetAttribute(SDL_GL_RED_SIZE,8);
	SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE,8);
	SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE,8);
	SDL_GL_SetAttribute(SDL_GL_ALPHA_SIZE,8);
	sdl_window=SDL_CreateWindow(title,wx,wy,xs,ys,
		SDL_WINDOW_OPENGL|SDL_WINDOW_SHOWN|(gl_windowed?SDL_WINDOW_RESIZABLE:SDL_WINDOW_FULLSCREEN));
	sdl_gl_context=SDL_GL_CreateContext(sdl_window);
	InitGL(xs,ys); // Sets xres,yres via ReSizeGLScene
	//SDL_WM_SetCaption(title,NULL); // SDL1
	memset(oldkeystate,0,keystate_m); memset(keystate,0,keystate_m);
	if (!gl_windowed) sjbshowcursor(0);
	return 1; // Success
}

#define KEY(x) (keystate[x])
#define KHIT(x) (keystate[x]&~oldkeystate[x])

void gfx_cleanup(void) {SDL_Quit();}

void gwarn(const char *message)
{
	gerrorlog(message);
	printf("\n%s\n",message);
}

void gerror(const char *message) {gwarn(message); gfx_cleanup(); exit(1);}

// Program for generating the following stuff
/*// gcc vmemtest.c -I ~/include -o vmemtest
#include <stdio.h>
#include <rnd/fgetline.c>
#include <rnd/streqi.c>

int main(void)
{
	FILE *in=fopen("/usr/include/SDL/SDL_keysym.h","rb");
	char *l,*s,*t;
	while (l=fgetline(in))
	{
		if (s=strstr(l,"SDLK_"))
		{
			if (t=strchr(s,'\t'))
			{
				*t=0;
				char *u; streqi(&u,s);
				for (t=u;*t;t++) if (*t>='a' && *t<='z') *t-=32;
				printf("#define VK_%s SDLK_%s\n",u+5,s+5);
				free(u);
			}
		}
		free(l);
	}
	return 0;
}*/

// Code autogenerated by replacements in the SDL_keyboard.h header
// Except it's not because SDL2 messed up the non-ASCII keycodes...
#define VK_BACKSPACE SDLK_BACKSPACE
#define VK_TAB SDLK_TAB
#define VK_CLEAR SDLK_CLEAR
#define VK_RETURN SDLK_RETURN
#define VK_PAUSE SDLK_PAUSE
#define VK_ESCAPE SDLK_ESCAPE
#define VK_SPACE SDLK_SPACE
#define VK_EXCLAIM SDLK_EXCLAIM
#define VK_QUOTEDBL SDLK_QUOTEDBL
#define VK_HASH SDLK_HASH
#define VK_DOLLAR SDLK_DOLLAR
#define VK_AMPERSAND SDLK_AMPERSAND
#define VK_QUOTE SDLK_QUOTE
#define VK_LEFTPAREN SDLK_LEFTPAREN
#define VK_RIGHTPAREN SDLK_RIGHTPAREN
#define VK_ASTERISK SDLK_ASTERISK
#define VK_PLUS SDLK_PLUS
#define VK_COMMA SDLK_COMMA
#define VK_MINUS SDLK_MINUS
#define VK_PERIOD SDLK_PERIOD
#define VK_SLASH SDLK_SLASH
#define VK_0 SDLK_0
#define VK_1 SDLK_1
#define VK_2 SDLK_2
#define VK_3 SDLK_3
#define VK_4 SDLK_4
#define VK_5 SDLK_5
#define VK_6 SDLK_6
#define VK_7 SDLK_7
#define VK_8 SDLK_8
#define VK_9 SDLK_9
#define VK_COLON SDLK_COLON
#define VK_SEMICOLON SDLK_SEMICOLON
#define VK_LESS SDLK_LESS
#define VK_EQUALS SDLK_EQUALS
#define VK_GREATER SDLK_GREATER
#define VK_QUESTION SDLK_QUESTION
#define VK_AT SDLK_AT
#define VK_LEFTBRACKET SDLK_LEFTBRACKET
#define VK_BACKSLASH SDLK_BACKSLASH
#define VK_RIGHTBRACKET SDLK_RIGHTBRACKET
#define VK_CARET SDLK_CARET
#define VK_UNDERSCORE SDLK_UNDERSCORE
#define VK_BACKQUOTE SDLK_BACKQUOTE
#define VK_A SDLK_a
#define VK_B SDLK_b
#define VK_C SDLK_c
#define VK_D SDLK_d
#define VK_E SDLK_e
#define VK_F SDLK_f
#define VK_G SDLK_g
#define VK_H SDLK_h
#define VK_I SDLK_i
#define VK_J SDLK_j
#define VK_K SDLK_k
#define VK_L SDLK_l
#define VK_M SDLK_m
#define VK_N SDLK_n
#define VK_O SDLK_o
#define VK_P SDLK_p
#define VK_Q SDLK_q
#define VK_R SDLK_r
#define VK_S SDLK_s
#define VK_T SDLK_t
#define VK_U SDLK_u
#define VK_V SDLK_v
#define VK_W SDLK_w
#define VK_X SDLK_x
#define VK_Y SDLK_y
#define VK_Z SDLK_z
#define VK_DELETE SDLK_DELETE
#define VK_KP0 (SDLK_KP0-0x40000000+0x80)
#define VK_KP1 (SDLK_KP1-0x40000000+0x80)
#define VK_KP2 (SDLK_KP2-0x40000000+0x80)
#define VK_KP3 (SDLK_KP3-0x40000000+0x80)
#define VK_KP4 (SDLK_KP4-0x40000000+0x80)
#define VK_KP5 (SDLK_KP5-0x40000000+0x80)
#define VK_KP6 (SDLK_KP6-0x40000000+0x80)
#define VK_KP7 (SDLK_KP7-0x40000000+0x80)
#define VK_KP8 (SDLK_KP8-0x40000000+0x80)
#define VK_KP9 (SDLK_KP9-0x40000000+0x80)
#define VK_KP_PERIOD (SDLK_KP_PERIOD-0x40000000+0x80)
#define VK_KP_DIVIDE (SDLK_KP_DIVIDE-0x40000000+0x80)
#define VK_KP_MULTIPLY (SDLK_KP_MULTIPLY-0x40000000+0x80)
#define VK_KP_MINUS (SDLK_KP_MINUS-0x40000000+0x80)
#define VK_KP_PLUS (SDLK_KP_PLUS-0x40000000+0x80)
#define VK_KP_ENTER (SDLK_KP_ENTER-0x40000000+0x80)
#define VK_KP_EQUALS (SDLK_KP_EQUALS-0x40000000+0x80)
#define VK_UP (SDLK_UP-0x40000000+0x80)
#define VK_DOWN (SDLK_DOWN-0x40000000+0x80)
#define VK_RIGHT (SDLK_RIGHT-0x40000000+0x80)
#define VK_LEFT (SDLK_LEFT-0x40000000+0x80)
#define VK_INSERT (SDLK_INSERT-0x40000000+0x80)
#define VK_HOME (SDLK_HOME-0x40000000+0x80)
#define VK_END (SDLK_END-0x40000000+0x80)
#define VK_PAGEUP (SDLK_PAGEUP-0x40000000+0x80)
#define VK_PAGEDOWN (SDLK_PAGEDOWN-0x40000000+0x80)
#define VK_F1 (SDLK_F1-0x40000000+0x80)
#define VK_F2 (SDLK_F2-0x40000000+0x80)
#define VK_F3 (SDLK_F3-0x40000000+0x80)
#define VK_F4 (SDLK_F4-0x40000000+0x80)
#define VK_F5 (SDLK_F5-0x40000000+0x80)
#define VK_F6 (SDLK_F6-0x40000000+0x80)
#define VK_F7 (SDLK_F7-0x40000000+0x80)
#define VK_F8 (SDLK_F8-0x40000000+0x80)
#define VK_F9 (SDLK_F9-0x40000000+0x80)
#define VK_F10 (SDLK_F10-0x40000000+0x80)
#define VK_F11 (SDLK_F11-0x40000000+0x80)
#define VK_F12 (SDLK_F12-0x40000000+0x80)
#define VK_F13 (SDLK_F13-0x40000000+0x80)
#define VK_F14 (SDLK_F14-0x40000000+0x80)
#define VK_F15 (SDLK_F15-0x40000000+0x80)
#define VK_NUMLOCK (SDLK_NUMLOCK-0x40000000+0x80)
#define VK_CAPSLOCK (SDLK_CAPSLOCK-0x40000000+0x80)
#define VK_SCROLLOCK (SDLK_SCROLLOCK-0x40000000+0x80)
#define VK_RSHIFT (SDLK_RSHIFT-0x40000000+0x80)
#define VK_LSHIFT (SDLK_LSHIFT-0x40000000+0x80)
#define VK_RCTRL (SDLK_RCTRL-0x40000000+0x80)
#define VK_LCTRL (SDLK_LCTRL-0x40000000+0x80)
#define VK_RALT (SDLK_RALT-0x40000000+0x80)
#define VK_LALT (SDLK_LALT-0x40000000+0x80)
#define VK_RMETA (SDLK_RMETA-0x40000000+0x80)
#define VK_LMETA (SDLK_LMETA-0x40000000+0x80)
#define VK_LSUPER (SDLK_LSUPER-0x40000000+0x80)
#define VK_RSUPER (SDLK_RSUPER-0x40000000+0x80)
#define VK_MODE (SDLK_MODE-0x40000000+0x80)
#define VK_COMPOSE (SDLK_COMPOSE-0x40000000+0x80)
#define VK_HELP (SDLK_HELP-0x40000000+0x80)
#define VK_PRINT (SDLK_PRINT-0x40000000+0x80)
#define VK_SYSREQ (SDLK_SYSREQ-0x40000000+0x80)
#define VK_BREAK (SDLK_BREAK-0x40000000+0x80)
#define VK_MENU (SDLK_MENU-0x40000000+0x80)
#define VK_POWER (SDLK_POWER-0x40000000+0x80)
#define VK_EURO (SDLK_EURO-0x40000000+0x80)
#define VK_UNDO (SDLK_UNDO-0x40000000+0x80)
// END autogenerated code

#define VK_ADD VK_KP_PLUS
#define VK_SUBTRACT VK_KP_MINUS
#define VK_PRIOR VK_PAGEUP
#define VK_NEXT VK_PAGEDOWN

// Ambidextrous versions
#define VK_SHIFT (keystate_m-1)
#define VK_CONTROL (keystate_m-2)
#define VK_ALT (keystate_m-3)

int sensiblekeycode(const int code)
{ // See https://wiki.libsdl.org/SDLKeycodeLookup
	if (code>=0x40000000) return code-0x40000000+0x80;
	else return code;
}

int dealwithmessages(void)
{ // Errr, looks like the Linux/SDL version doesn't poll the mouse position or buttons yet
	SDL_Event event;
	memcpy(oldkeystate,keystate,keystate_m);
	omb=mb;
	mb&=~(4|8); // Reset wheel events
	int mouserel=SDL_GetRelativeMouseMode();
	while (SDL_PollEvent(&event)) switch (event.type)
	{
		case SDL_KEYDOWN: keystate[sensiblekeycode(event.key.keysym.sym)]=1; break;
		case SDL_KEYUP: keystate[sensiblekeycode(event.key.keysym.sym)]=0; break;
		case SDL_MOUSEMOTION:
			if (mouserel) {mx+=event.motion.xrel; my+=event.motion.yrel;}
			else {mx=event.motion.x; my=event.motion.y;}
			break;
		case SDL_MOUSEBUTTONDOWN: case SDL_MOUSEBUTTONUP:
		{
			int m=(event.button.button==SDL_BUTTON_LEFT?1:
					event.button.button==SDL_BUTTON_RIGHT?2:
					//event.button.button==SDL_BUTTON_WHEELUP?4: // "Older" SDL way of doing it: wheel is buttons(?)
					//event.button.button==SDL_BUTTON_WHEELDOWN?8:
					event.button.button==SDL_BUTTON_MIDDLE?16:
					event.button.button==SDL_BUTTON_X1?32:64),
				b=(event.button.state==SDL_PRESSED || m==4 || m==8);
				mb=(mb&~m)|(b*m);
		} break;
		case SDL_MOUSEWHEEL: // Requires SDL2, which does indeed come with Debian 11
		{ // Needs mb&=~(4|8); at top
			if (event.wheel.y>0) mb|=4; // Scroll up
			else if (event.wheel.y<0) mb|=8; // Scroll down
		} break;
		case SDL_WINDOWEVENT: switch (event.window.event)
		{ // See https://wiki.libsdl.org/SDL_WindowEvent
			case SDL_WINDOWEVENT_SIZE_CHANGED:
				ReSizeGLScene(event.window.data1,event.window.data2); // Has to be called in thread that actually has OpenGL in it
				break; // Note that this clears the depth buffer to 0.0 not 1.0 so if plotting in immediate mode, need to glClear(GL_DEPTH_BUFFER_BIT);
		} break;
		case SDL_QUIT: gfx_cleanup(); return 1; break;
	}
	keystate[VK_SHIFT]=keystate[VK_LSHIFT]|keystate[VK_RSHIFT];
	keystate[VK_CONTROL]=keystate[VK_LCTRL]|keystate[VK_RCTRL];
	keystate[VK_ALT]=keystate[VK_LALT]|keystate[VK_RALT];
	return 0;
}

#else // Not linux therefore Windows

static HGLRC hRC;		// Permanent Rendering Context
static HDC hDC;		// Private GDI Device Context
HWND winhandle;

int mb_real,mb_stuck=1;

#define GRAPHICS(X,Y,T) if (!makethewindow(X,Y,T,hinstance)) return;

#define KEY(x) (keystate[x]>>7)
#define KHIT(x) ((keystate[x]&~oldkeystate[x])>>7)

void gfx_cleanup(void)
{ // Used to be called "gracefulexit" but no longer exits so program-specific exit stuff can happen too
	if (!gl_windowed) ChangeDisplaySettings(NULL,0);
	wglMakeCurrent(hDC,NULL);
	wglDeleteContext(hRC);
	ReleaseDC(winhandle,hDC);
	//PostQuitMessage(0);
	//exit(0);
}

void gwarn(const char *message)
{
	gerrorlog(message);
	MessageBox(NULL,message,"Errr",MB_ICONERROR|MB_SETFOREGROUND|MB_OK);
}

void gerror(const char *message)
{
	gerrorlog(message);
	MessageBox(NULL,message,"Oh sod it",MB_ICONERROR|MB_SETFOREGROUND|MB_OK);
	gfx_cleanup(); exit(1);
}

int dealwithmessages_leave=0;

int dealwithmessages(void)
{
	omb=mb;
	mb&=~(4|8); // Reset wheel events
	MSG msg; // Process all messages
	while (PeekMessage(&msg,NULL,0,0,PM_NOREMOVE))
	{
		if (GetMessage(&msg,NULL,0,0))
		{
			TranslateMessage(&msg); // 161=WM_NCLBUTTONDOWN, non-client area left button down
			DispatchMessage(&msg); // <-- Multiple seconds disappear in one of these two while dragging, on msg.message=161
		}
		else return dealwithmessages_leave=1; // WM_QUIT(?)
	}
	static char closing=0;
	if (!IsIconic(winhandle))
	{
		if (!closing && !gl_windowed) SetWindowPos(winhandle,HWND_TOPMOST,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE); // Always on top (stops Miranda IM appearing on top for instance)
		memcpy(oldkeystate,keystate,keystate_m);
		GetKeyboardState(keystate);
		if (KEY(VK_ALT) && KEY(VK_TAB)) closing=1;
		if (closing) CloseWindow(winhandle);
	}
	else closing=0;
	return dealwithmessages_leave;
}

/*unsigned DefWindowProc_t(MSG *m)
{
	DefWindowProc(m->hwnd,m->message,m->wParam,m->lParam);
	free(m);
	return 0;
}*/

void (*wm_gesture_handler)(WPARAM,LPARAM)=NULL;

LRESULT CALLBACK windowproc(HWND hWnd,UINT message,WPARAM wParam,LPARAM lParam)
{
	switch (message)
	{
		case WM_DESTROY: case WM_CLOSE: // Trigger using SendMessage(winhandle,WM_CLOSE,0,0);
			gfx_cleanup(); dealwithmessages_leave=1; break; // Seems these messages are only picked up here not in dealwithmessages(), so communicate with global
		//case WM_KEYDOWN: keystate[wParam]=TRUE; break;
		//case WM_KEYUP: keystate[wParam]=FALSE; break; // Commented out because I've implemented this in dealwithmessages()
		case WM_SYSKEYDOWN: return 0; // Stop ALT or F10 activating the window "menu"
		case WM_CHAR:
		{ // Buffering for ordered text input
			//keyhit=1; // Came from <gfx_dx.c>
			int l=strlen(gotch);
			if (l<499)
			{
				gotch[l+1]=0;
				gotch[l]=(char)wParam;
			}
			break;
		}
		case WM_SIZE:
			ReSizeGLScene(LOWORD(lParam),HIWORD(lParam)); // Has to be called in thread that actually has OpenGL in it
			break; // Note that this clears the depth buffer to 0.0 not 1.0 so if plotting in immediate mode, need to glClear(GL_DEPTH_BUFFER_BIT);
		/*case WM_SIZING:
		{
			RECT *pr=(RECT *)lParam;
			ReSizeGLScene(pr->right-pr->left,pr->bottom-pr->top);
		} break;*/
		case WM_MOUSEMOVE:
		case WM_LBUTTONDOWN: case WM_LBUTTONUP: //case WM_LBUTTONDBLCLK:
		case WM_RBUTTONDOWN: case WM_RBUTTONUP: //case WM_RBUTTONDBLCLK:
		case WM_MBUTTONDOWN: case WM_MBUTTONUP: //case WM_MBUTTONDBLCLK:
			mx=(int)LOWORD(lParam);
			my=(int)HIWORD(lParam);
			mb_real=(int)wParam&(MK_LBUTTON|MK_RBUTTON|MK_MBUTTON);
			if (!mb_real) mb_stuck=0;
			if (!mb_stuck) mb=(mb&12)|(mb_real&~12);
			break;
		case WM_MOUSEWHEEL:
			if ((int)wParam>0) mb|=4; else mb|=8;
			break;
		case 0x0119: //WM_GESTURE:
			if (wm_gesture_handler) wm_gesture_handler(wParam,lParam);
			break;
		/*case WM_NCLBUTTONDOWN:
		{ // Call DefWindowProc in its own thread when it does a blocking (modal) operation, e.g. window resizing
			MSG *m=malloc(sizeof(MSG)); // This doesn't exactly work, still manages to block somehow
			m->hwnd=hWnd; m->message=message; // I could try putting the WHOLE of dealwithmessages in a thread?
			m->wParam=wParam; m->lParam=lParam;
			unsigned long mallards;
			CloseHandle(CreateThread(NULL,0,DefWindowProc_t,m,0,&mallards));
			return 0;
		}*/
		default:
			return (DefWindowProc(hWnd,message,wParam,lParam));
	}
	return 0;
}

HICON makethewindow_icon=NULL;

#include <rnd/enumdisplays.c>
#include <rnd/sjbshowcursor.c>

int makethewindow(const int xs,const int ys,const char *title,const HINSTANCE hinstance)
{ // Returns 1 on success, 0 on failure
	//SetProcessDPIAware(); isn't in TDM-GCC's Windows headers, get from DLL...
	{
		HMODULE lib=LoadLibrary("user32.dll");
		if (lib)
		{
			BOOL (*SetProcessDPIAware)(void)=(BOOL (*)(void))GetProcAddress(lib,"SetProcessDPIAware");
			if (SetProcessDPIAware) SetProcessDPIAware();
		    FreeLibrary(lib);
		}
	}
	int wx=0,wy=0;
	if (xs==0) // Autodetect
	{
		if (ys==0) // Primary display
		{
			xres=GetSystemMetrics(SM_CXSCREEN);
			yres=GetSystemMetrics(SM_CYSCREEN);
		}
		else // Display number ys (starting from 1)
		{
			LSet ds=enumdisplays();
			RECT *r=(RECT *)ds.a;
			int i=(ys>=1 && ys-1<ds.m?ys-1:0);
			wx=r[i].left;
			wy=r[i].top;
			xres=r[i].right-r[i].left;
			yres=r[i].bottom-r[i].top;
			LSet_free(&ds);
		}
		if (gl_windowed==1) {yres=(int)(yres*0.6); xres=yres*4/3;}
	}
	else if (xs==-1) // Multi-monitor spanning mode
	{
		if (ys<=0) // All monitors
		{
			wx=GetSystemMetrics(SM_XVIRTUALSCREEN);
			wy=GetSystemMetrics(SM_YVIRTUALSCREEN);
			xres=GetSystemMetrics(SM_CXVIRTUALSCREEN);
			yres=GetSystemMetrics(SM_CYVIRTUALSCREEN);
		}
		else // Span subset of monitors with bit flags in ys
		{
			LSet ds=enumdisplays();
			RECT *r=(RECT *)ds.a;
			wx=wy=999999999; int vx=-wx,vy=-wy;
			for (int n=0;n<31 && n<ds.m;n++) if (ys&(1<<n))
			{
				if (r[n].left<wx) wx=r[n].left;
				if (r[n].top<wy) wy=r[n].top;
				if (r[n].right>vx) vx=r[n].right;
				if (r[n].bottom>vy) vy=r[n].bottom;
			}
			LSet_free(&ds);
			xres=vx-wx;
			yres=vy-wy;
		}
	}
	else {xres=xs; yres=ys;}
	int wxs=xres,wys=yres;
	wx+=makethewindow_shiftx; wy+=makethewindow_shifty;
	if (gl_windowed==1)
	{ // Try and make it centred on the mouse pointer but then clipped to the span of the screens
		wxs+=2*GetSystemMetrics(SM_CXFRAME); wys+=GetSystemMetrics(SM_CYCAPTION)+2*GetSystemMetrics(SM_CYFRAME);
		POINT mp; GetCursorPos(&mp);
		wx=Maxi(GetSystemMetrics(SM_XVIRTUALSCREEN),Mini(mp.x-wxs/2,GetSystemMetrics(SM_XVIRTUALSCREEN)+GetSystemMetrics(SM_CXVIRTUALSCREEN)-wxs));
		wy=Maxi(GetSystemMetrics(SM_YVIRTUALSCREEN),Mini(mp.y-wys/2,GetSystemMetrics(SM_YVIRTUALSCREEN)+GetSystemMetrics(SM_CYVIRTUALSCREEN)-wys));
	}
	if (xres==0 || yres==0) return 0;
	memset(keystate,0,keystate_m); memset(oldkeystate,0,keystate_m);
	WNDCLASS wc; // Windows Class Structure Used To Set Up The Type Of Window
	wc.style=CS_HREDRAW|CS_VREDRAW|CS_OWNDC;
	wc.lpfnWndProc=(WNDPROC)windowproc;
	wc.cbClsExtra=0;
	wc.cbWndExtra=0;
	wc.hInstance=hinstance;
	wc.hIcon=makethewindow_icon;
	wc.hCursor=LoadCursor(NULL,IDC_ARROW);
	wc.hbrBackground=NULL;
	wc.lpszMenuName=NULL;
	wc.lpszClassName="OpenGL WinClass";
	if (!RegisterClass(&wc)) reportfatal("Failed To Register The Window Class.");
	winhandle=CreateWindow("OpenGL WinClass",title,
		WS_VISIBLE|WS_CLIPCHILDREN|WS_CLIPSIBLINGS|
		(gl_windowed==1?WS_TILED|WS_SYSMENU|WS_THICKFRAME|WS_MINIMIZEBOX|WS_MAXIMIZEBOX:WS_POPUP),
		wx,wy,wxs,wys,NULL,NULL,hinstance,NULL);
	if (!winhandle) reportfatal("Window Creation Error.");
	const static PIXELFORMATDESCRIPTOR pfd=
	{
		sizeof(PIXELFORMATDESCRIPTOR),		// Size Of This Pixel Format Descriptor
		1,									// Version Number (?)
		PFD_DRAW_TO_WINDOW				// Format Must Support Window
		| PFD_SUPPORT_OPENGL				// Format Must Support OpenGL
		#ifndef SBCLIB_GRAPHICS_IMMEDIATE
		| PFD_DOUBLEBUFFER					// Must Support Double Buffering
		#endif
		,PFD_TYPE_RGBA,						// Request An RGBA Format
		24,									// Select A 16M Color Depth (changed from 16 in tutorial)
		0, 0, 0, 0, 0, 0,					// Color Bits Ignored (?)
		0,									// No Alpha Buffer
		0,									// Shift Bit Ignored (?)
		0,									// No Accumulation Buffer
		0, 0, 0, 0,							// Accumulation Bits Ignored (?)
		32,									// Highest bitness Z-Buffer available (Depth Buffer)
		0,									// No Stencil Buffer
		0,									// No Auxiliary Buffer (?)
		PFD_MAIN_PLANE,						// Main Drawing Layer
		0,									// Reserved (?)
		0, 0, 0								// Layer Masks Ignored (?)
	};
	hDC=GetDC(winhandle);				// Gets A Device Context For The Window
	int PixelFormat=ChoosePixelFormat(hDC,&pfd);		// Finds The Closest Match To The Pixel Format We Set Above
	if (!PixelFormat) reportfatal("Can't Find A Suitable PixelFormat.");
	if (!SetPixelFormat(hDC,PixelFormat,&pfd)) reportfatal("Can't Set The PixelFormat.");
	hRC=wglCreateContext(hDC);
	if (!hRC) reportfatal("Can't Create A GL Rendering Context.");
	if (!wglMakeCurrent(hDC,hRC)) reportfatal("Can't activate GLRC.");
	InitGL(xres,yres);
	if (xs>0 && !gl_windowed && makethewindow_shiftx==0 && makethewindow_shifty==0)
	{
		DEVMODE dmScreenSettings;
		memset(&dmScreenSettings,0,sizeof(DEVMODE));
		dmScreenSettings.dmSize=sizeof(DEVMODE);
		dmScreenSettings.dmPelsWidth=xres; // Screen Width
		dmScreenSettings.dmPelsHeight=yres; // Screen Height
		dmScreenSettings.dmFields=DM_PELSWIDTH|DM_PELSHEIGHT; // Pixel Mode
		ChangeDisplaySettings(&dmScreenSettings,CDS_FULLSCREEN); // Switch To Full Screen
	}
	if (!gl_windowed) sjbshowcursor(0);
	return 1; // Success
}
#endif

#define SBCLIB_GFX_OPENGL
#endif
