#ifndef SBCLIB_GL_SHADERPROCS
#include <graphics.c>
//#include <gl/glext.h>

typedef char GLchar;

GLuint (APIENTRY *glCreateShader)(GLenum shaderType);
void (APIENTRY *glShaderSource)(GLuint shader, int numOfStrings, const char **strings, int *lenOfStrings);
void (APIENTRY *glCompileShader)(GLuint shader);

GLuint (APIENTRY *glCreateProgram)(void);
void (APIENTRY *glAttachShader)(GLuint program, GLuint shader);
void (APIENTRY *glLinkProgram)(GLuint program);
void (APIENTRY *glUseProgram)(GLuint prog);

GLint (APIENTRY *glGetUniformLocation)(GLuint program, const char *name);
void (APIENTRY *glUniform1f)(GLint location, GLfloat v0);
void (APIENTRY *glUniform2f)(GLint location, GLfloat v0, GLfloat v1);
void (APIENTRY *glUniform3f)(GLint location, GLfloat v0, GLfloat v1, GLfloat v2);
void (APIENTRY *glUniform1i)(GLint location, GLint v0);
void (APIENTRY *glUniform2i)(GLint location, GLint v0, GLint v1);
void (APIENTRY *glUniform1fv)(GLint location, GLsizei count, const GLfloat *value);
void (APIENTRY *glUniform2fv)(GLint location, GLsizei count, const GLfloat *value);
void (APIENTRY *glUniform3fv)(GLint location, GLsizei count, const GLfloat *value);
void (APIENTRY *glUniform1iv)(GLint location, GLsizei count, const GLint *value);
void (APIENTRY *glUniformMatrix4fv)(GLint location, GLsizei count, GLboolean transpose, const GLfloat *value);
  	
void (APIENTRY *glGetShaderiv)(GLuint shader, GLenum pname, GLint *params);
void (APIENTRY *glGetShaderInfoLog)(GLuint shader, GLsizei maxLength, GLsizei *length, GLchar *infoLog);
void (APIENTRY *glGetProgramiv)(GLuint program, GLenum pname, GLint *params);
void (APIENTRY *glGetProgramInfoLog)(GLuint program, GLsizei maxLength, GLsizei *length, GLchar *infoLog);

#ifndef __linux__
void (APIENTRY *glActiveTexture)(GLenum texture);
#endif

void initshaderprocs(void)
{
	static char done=0;
	if (done) return;
	if (glversion()<2.0)
	{
		char str[100];
		sprintf(str,"OpenGL version 2.0 or higher required for shaders; you have version %g",glversion());
		reportfatala(str);
	}
	#define D(X) *(void **)&X=sjbglGetProcAddress(#X);
	D(glCreateShader); D(glShaderSource); D(glCompileShader);
	D(glCreateProgram) D(glAttachShader) D(glLinkProgram) D(glUseProgram)
	D(glGetUniformLocation) D(glUniform1f) D(glUniform2f) D(glUniform3f) D(glUniform1i) D(glUniform2i)
	D(glUniform1fv) D(glUniform2fv) D(glUniform3fv) D(glUniform1iv)
	D(glUniformMatrix4fv)
	D(glGetShaderiv) D(glGetShaderInfoLog) D(glGetProgramiv) D(glGetProgramInfoLog)
	#ifndef __linux__
	D(glActiveTexture)
	#endif
	#undef D
	done=1;
}

int hasshaderprocs(void)
{
	static int done=0,ret;
	if (done) return ret;
	done=1;
	if (glversion()<2.0) return ret=0;
	else {initshaderprocs(); return ret=1;}
}

GLuint shaderfromcode(const GLenum type,const char *code)
{ // Creates and compiles a shader from a single code string
	GLuint ret=glCreateShader(type);
	glShaderSource(ret,1,&code,NULL);
	glCompileShader(ret);
	return ret;
}

GLuint shaderfromcodelibs(const GLenum type,char **libs,const int m)
{ // Creates and compiles a shader from a sequence of m code strings ("libraries")
	GLuint ret=glCreateShader(type);
	glShaderSource(ret,m,(const char **)libs,NULL);
	glCompileShader(ret);
	return ret;
}

#include <rnd/streqi.c>

char *shaderlog(const GLuint id,const char isprog=0)
{ // String constructor, gets log information from shader or program objects
	int len; char *ret;
	if (isprog) glGetProgramiv(id,GL_INFO_LOG_LENGTH,&len);
	else glGetShaderiv(id,GL_INFO_LOG_LENGTH,&len);
	if (len)
	{
		ret=(char *)malloc(len);
		if (isprog) glGetProgramInfoLog(id,len,NULL,ret);
		else glGetShaderInfoLog(id,len,NULL,ret);
		return ret;
	}
	else {streqi(&ret,"No information log"); return ret;}
}

char *shaderlogfile=(char *)"shader_error.log";

void shaderlogfile_check(const GLuint id,const char isprog=0)
{ // Output to log file only if there are errors
	int success;
	if (isprog) glGetProgramiv(id,GL_LINK_STATUS,&success);
	else glGetShaderiv(id,GL_COMPILE_STATUS,&success);
	if (!success)
	{
		char *e=shaderlog(id,isprog);
		FILE *out=fopen(shaderlogfile,"at");
		fprintf(out,"%s\n",e);
		fclose(out);
		free(e);
	}
}

GLuint shaderprogvf(const char *vcode,const char *fcode,char **vlog=NULL,char **flog=NULL,char **plog=NULL)
{ // If xlog is not NULL, *xlog becomes a constructed string, free with free(*xlog)
	if (!hasshaderprocs()) return 0;
	GLuint ret=glCreateProgram(),vertsh,fragsh;
	if (vcode)
	{
		vertsh=shaderfromcode(GL_VERTEX_SHADER,vcode);
		if (vlog) *vlog=shaderlog(vertsh);
		if (shaderlogfile) shaderlogfile_check(vertsh);
		glAttachShader(ret,vertsh);
	}
	if (fcode)
	{
		fragsh=shaderfromcode(GL_FRAGMENT_SHADER,fcode);
		if (flog) *flog=shaderlog(fragsh);
		if (shaderlogfile) shaderlogfile_check(fragsh);
		glAttachShader(ret,fragsh);
	}
	glLinkProgram(ret);
	if (plog) *plog=shaderlog(ret,1);
	if (shaderlogfile) shaderlogfile_check(ret,1);
	return ret;
}

GLuint shaderprogvflibs(char **vlibs,const int vm,
	char **flibs,const int fm,char **vlog=NULL,char **flog=NULL,char **plog=NULL)
{ // If xlog is not NULL, *xlog becomes a constructed string, free with free(*xlog)
	if (!hasshaderprocs()) return 0;
	GLuint ret=glCreateProgram(),vertsh,fragsh;
	if (vlibs)
	{
		vertsh=shaderfromcodelibs(GL_VERTEX_SHADER,vlibs,vm);
		if (vlog) *vlog=shaderlog(vertsh);
		if (shaderlogfile) shaderlogfile_check(vertsh);
		glAttachShader(ret,vertsh);
	}
	if (flibs)
	{
		fragsh=shaderfromcodelibs(GL_FRAGMENT_SHADER,flibs,fm);
		if (flog) *flog=shaderlog(fragsh);
		if (shaderlogfile) shaderlogfile_check(fragsh);
		glAttachShader(ret,fragsh);
	}
	glLinkProgram(ret);
	if (plog) *plog=shaderlog(ret,1);
	if (shaderlogfile) shaderlogfile_check(ret,1);
	return ret;
}

GLuint shaderprogfrag(const char *code,char **slog=NULL,char **plog=NULL)
{ // Fragment shader only, plus returns zero if no shaders available (also no init required)
	if (!hasshaderprocs()) return 0;
	return shaderprogvf(NULL,code,NULL,slog,plog);
}

GLuint shaderprogfraglibs(char **libs,const int m,char **slog=NULL,char **plog=NULL)
{ // Fragment shader only, plus returns zero if no shaders available (also no init required)
	if (!hasshaderprocs()) return 0;
	return shaderprogvflibs(NULL,0,libs,m,NULL,slog,plog);
}

/* No shaderprogvert because it doesn't really work without a minimal fragment shader like:
void main()
{
	gl_FragColor=gl_Color; // (Set gl_FrontColor e.g. =gl_Color; in vertex shader)
}
*/
#define SBCLIB_GL_SHADERPROCS
#endif
